@@ -2,7 +2,8 @@ import logging
 import signal
 import sys
 import os.path
-from datetime import datetime
+from datetime import datetime, timedelta
+from time import sleep
 from optparse import make_option, OptionParser
 
 from setproctitle import setproctitle
@@ -35,6 +36,9 @@ class Worker(object):
     max_loops = 1000
     # max delay for blpop
     timeout = 30
+    # minimum time in seconds between the update of keys to fetch
+    fetch_priorities_delay = 25  # by default < timeout to be sure to
+                                 # fetch them after a timeout occured
 
     # we want to intercept SIGTERM and SIGINT signals, to stop gracefuly
     terminate_gracefuly = True
@@ -42,7 +46,8 @@ class Worker(object):
     def __init__(self, name=None, callback=None,
                  queue_model=None, job_model=None, error_model=None,
                  logger_base_name=None, logger_level=None, save_errors=None,
-                 max_loops=None, terminate_gracefuly=None, timeout=None):
+                 max_loops=None, terminate_gracefuly=None, timeout=None,
+                 fetch_priorities_delay=None):
         """
         Create the worker by saving arguments, doing some checks, preparing
         logger and signals management, and getting queues keys.
@@ -74,6 +79,8 @@ class Worker(object):
             self.save_errors = save_errors
         if timeout is not None:
             self.timeout = timeout
+        if fetch_priorities_delay is not None:
+            self.fetch_priorities_delay = fetch_priorities_delay
 
         # prepare logging
         if logger_base_name is not None:
@@ -90,7 +97,9 @@ class Worker(object):
         self.end_forced = False  # set it to True in "execute" to force stop just after
         self.status = None  # is set to None/waiting/running by the worker
         self.end_signal_caught = False  # internaly set to True if end signal caught
-        self.update_callbacks = []  # callbacks to call when status is updated
+        self.last_fetch_priorities = None  # last time the keys to fetch were updated
+
+        self._update_status_callbacks = []  # callbacks to call when status is updated
 
     @staticmethod
     def _assert_correct_model(model_to_check, model_reference, obj_name):
@@ -107,8 +116,21 @@ class Worker(object):
         """
         Catch some system signals to handle them internaly
         """
-        signal.signal(signal.SIGTERM, self.catch_end_signal)
-        signal.signal(signal.SIGINT, self.catch_end_signal)
+        try:
+            signal.signal(signal.SIGTERM, self.catch_end_signal)
+            signal.signal(signal.SIGINT, self.catch_end_signal)
+        except ValueError:
+            self.log('Signals cannot be caught in a Thread', level='warning')
+
+    def stop_handling_end_signal(self):
+        """
+        Stop handling the SIGINT and SIGTERM signals
+        """
+        try:
+            signal.signal(signal.SIGTERM, signal.SIG_DFL)
+            signal.signal(signal.SIGINT, signal.SIG_DFL)
+        except ValueError:
+            self.log('Signals cannot be caught in a Thread', level='warning')
 
     def set_logger(self):
         """
@@ -151,20 +173,20 @@ class Worker(object):
         Save the new status and call all defined callbacks
         """
         self.status = status
-        for callback in self.update_callbacks:
+        for callback in self._update_status_callbacks:
             callback(self)
 
-    def add_update_callback(self, callback):
+    def _add_update_status_callback(self, callback):
         """
         Add a callback to call when the status is updated
         """
-        self.update_callbacks.append(callback)
+        self._update_status_callbacks.append(callback)
 
-    def remove_update_callback(self, callback):
+    def _remove_update_status_callback(self, callback):
         """
         Remove a callback from ones set to be called when the status is updated
         """
-        self.update_callbacks.remove(callback)
+        self._update_status_callbacks.remove(callback)
 
     def wait_for_job(self):
         """
@@ -198,6 +220,10 @@ class Worker(object):
         When a SIGINT/SIGTERM signal is caught, this method is called, asking
         for the worker to terminate as soon as possible.
         """
+        if self.end_signal_caught:
+            self.log('Previous signal caught, will end soon')
+            return
+
         signal_name = dict((getattr(signal, n), n) for n in dir(signal)
                         if n.startswith('SIG') and '_' not in n).get(signum, signum)
 
@@ -205,8 +231,9 @@ class Worker(object):
             self.log('Catched %s signal: stopping after current job' % signal_name,
                      level='critical')
         else:
-            self.log('Catched %s signal: stopping right now' % signal_name,
-                     level='critical')
+            delay = self.timeout if self.status == 'waiting' else self.fetch_priorities_delay
+            self.log('Catched %s signal: stopping in max %d seconds' % (
+                     signal_name, delay), level='critical')
 
         self.end_signal_caught = self.end_forced = True
 
@@ -221,9 +248,10 @@ class Worker(object):
 
     def update_keys(self):
         """
-        Update the redis keys to listen for new jobs.
+        Update the redis keys to listen for new jobs priorities.
         """
         self.keys = self.queue_model.get_keys(self.name)
+        self.last_update_keys = datetime.utcnow()
 
     def count_waiting_jobs(self):
         """
@@ -250,16 +278,38 @@ class Worker(object):
 
         self.set_status('starting')
 
+        must_stop = False
+
+        # get keys or wait for queues available if no ones are yet
         self.update_keys()
         if not self.keys:
-            self.log('No queues with the name %s.' % self.name, level='error')
-            self.set_status('aborted')
-            return
+            self.log('No queues yet with the name %s.' % self.name, level='warning')
+            while not self.keys:
+                sleep(self.fetch_priorities_delay)
+                must_stop = self.must_stop()
+                if must_stop:
+                    break
+                self.update_keys()
 
-        self.run_started()
+        if not must_stop:
+            self.run_started()
+            self._main_loop()
+
+        self.set_status('terminated')
+        self.run_ended()
+        if self.terminate_gracefuly:
+            self.stop_handling_end_signal()
+
+    def _main_loop(self):
+        """
+        Run jobs until must_stop returns True
+        """
+        fetch_priorities_delay = timedelta(seconds=self.fetch_priorities_delay)
 
         while not self.must_stop():
             self.set_status('waiting')
+            if self.last_update_keys + fetch_priorities_delay < datetime.utcnow():
+                self.update_keys()
             try:
                 queue_and_job = self.wait_for_job()
                 if queue_and_job is None:
@@ -293,9 +343,6 @@ class Worker(object):
                     self.log('[%s] unexpected error: %s' % (identifier, str(e)),
                              level='error')
 
-        self.set_status('terminated')
-        self.run_ended()
-
     def run_ended(self):
         """
         Called just after ending the run loop. Actually only do logging.
@@ -438,6 +485,9 @@ class WorkerConfig(object):
         make_option('--timeout', type='int', dest='timeout',
             help='Max delay (seconds) to wait for a redis BLPOP call (0 for no timeout), e.g. --timeout=30'),
 
+        make_option('--fetch-priorities-delay', type='int', dest='fetch_priorities_delay',
+            help='Min delay (seconds) to wait before fetching new priority queues, e.g. --fetch-priorities-delay=30'),
+
         make_option('--database', action='store', dest='database',
             help='Redis database to use (host:port:db), e.g. --database=localhost:6379:15'),
 
@@ -455,7 +505,8 @@ class WorkerConfig(object):
 
     worker_options = ('name', 'job_model', 'queue_model', 'error_model',
                       'callback', 'logger_base_name', 'logger_level',
-                      'save_errors', 'max_loops', 'terminate_gracefuly', 'timeout')
+                      'save_errors', 'max_loops', 'terminate_gracefuly',
+                      'timeout', 'fetch_priorities_delay')
 
     @staticmethod
     def _import_module(module_uri):
@@ -555,6 +606,9 @@ class WorkerConfig(object):
         if self.options.timeout is not None and self.options.timeout < 0:
             self.parser.error('The timeout argument (%s) must be a positive integer (including 0)' % self.options.timeout)
 
+        if self.options.fetch_priorities_delay is not None and self.options.fetch_priorities_delay <= 0:
+            self.parser.error('The fetch_priorities_delay argument (%s) must be a positive integer' % self.options.fetch_priorities_delay)
+
         self.database_config = None
         if self.options.database:
             host, port, db = self.options.database.split(':')
@@ -650,7 +704,7 @@ class WorkerConfig(object):
         worker_options = self.prepare_worker_options()
         self.worker = self.options.worker_class(**worker_options)
         if self.update_title:
-            self.worker.add_update_callback(self.update_proc_title)
+            self.worker._add_update_status_callback(self.update_proc_title)
             self.update_proc_title()
 
         if not self.worker.logger.handlers:
